use std::io;
use libbitfyre::base;

use crate::{diff::{self, Compare}, eval::{eval, EvalError, EvalResult}, map::{input_error, MapDisplay}};

const NAME_VERSION: &str = concat!(env!("CARGO_PKG_NAME"), " ", env!("CARGO_PKG_VERSION"));

macro_rules! printlnlazy {
	($($l:literal)+) => {
		$(println!($l);)*
	};
}

fn help_docs() {
	printlnlazy!(
		"hi"
		"hi2"
		"hi3"
	);
}

fn init_message() {
	println!("{NAME_VERSION}");
}

pub struct Out;
impl Out {
	fn error(&self, eval_error: EvalError) {
		match eval_error {
			EvalError::I64Conversion(parse_int_error) => println!("{parse_int_error}"),
			EvalError::InvalidOperator(invalid_op) => println!("The operator {:?} is invalid.", invalid_op),
		}
	}

	pub fn evaled(&self, eval_out: Result<EvalResult, EvalError>) {
		eval_out.map_or_else(|e| self.error(e), |result| {
			let eval_base2 = base::binary(result.eval);
			let left       = base::binary(result.input.left);
			let right      = base::binary(result.input.right);

			let mut differ = diff::Comparer::new();
			let diffed = differ.compare(Compare {
				right: right.clone(),
				left: left.clone(),
				eval: eval_base2.clone(),
			});

			println!("{}", diffed.left);
			println!("{}", diffed.right);
			println!("{}", "-".repeat(eval_base2.chars().count()));
			println!("{}", diffed.eval)
		})
	}
}

pub struct Bitfyre(bool); //Terminate status
impl Bitfyre {
	pub const fn new() -> Self {
		Self(false)
	}

	//<E: fmt::Display> term_Reason: Option<E>
	pub fn terminate(&mut self, term_reason: Option<io::Error>) {
		term_reason.map_or_else(|| println!("bitfyre: Received exit signal."), input_error);
		self.0 = true
	}

	fn listener_stdin(&mut self) -> Result<(), io::Error> {
		io::Write::flush(&mut io::stdout()).map(|()| {
			let mut input = String::new();
			io::stdin().read_line(&mut input).map_or_display(|_size| match input.trim() {
				"exit" => self.terminate(None),
				"help" => help_docs(),
				trim => Out.evaled(eval(trim.split_whitespace().collect()))
			})
		})
	}

	pub fn listen(&mut self) {
		init_message();

		loop {
			if self.0 { break } else {
				print!("> ");
				if let Err(flush_err) = self.listener_stdin() {
					self.terminate(Some(flush_err))
				}
			}
		}
	}
}